'''

This is REALLY messy and needs attending too.. 20160310
Be able to pass in dictionary of params
be run score method outside of the training.


'''
from __future__ import print_function
import numpy as np
import pandas as pd


from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import Imputer, StandardScaler
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.decomposition import PCA
from sklearn import cross_validation
from sklearn.metrics import classification_report
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis



class NetworkClassifer():

    def __init__(self, features, labels, validation_features, validation_labels):
        self.features = features
        self.feature_labels = ['min','max','mean','skew','std','kurtosis','sum of absolute difference','baseline_n',
                               'baseline_diff','baseline_diff_skew','n_pks','n_vals','av_pk','av_val','av pk val range',
                               '1 hz','5 hz','10 hz','15 hz','20 hz','30 hz','60 hz','90 hz']

        self.labels = np.ravel(labels)
        self.validation_features = validation_features
        self.validation_labels = np.ravel(validation_labels)
        self.impute_and_scale()

    def impute_and_scale(self):
        print('Scaling and imputing training dataset...')
        self.imputer = Imputer(missing_values='NaN', strategy='mean', axis=0)
        self.imputer.fit(self.features)
        imputed_features = self.imputer.transform(self.features)
        self.std_scaler = StandardScaler()
        self.std_scaler.fit(imputed_features)
        self.iss_features = self.std_scaler.transform(imputed_features)
        print('Done')

        print('Scaling and imputing validation features using training dataset...')
        imputed_validation_features = self.imputer.transform(self.validation_features)
        self.iss_validation_features = self.std_scaler.transform(imputed_validation_features)
        print('Done')

    def _cross_validation(self,clf, k_folds = 5):
        self.scores = cross_validation.cross_val_score(clf, self.iss_features, self.labels, cv=k_folds,n_jobs=5, scoring = 'roc_auc')

    def randomforest_info(self, max_trees = 1000, step = 40, k_folds = 5):
        print('Characterising R_forest. Looping through trees: ')
        self.treedata = np.zeros((max_trees/step, 10))
        for i,n_trees in enumerate(np.arange(0, max_trees,step)):
            if n_trees == 0:
                n_trees = 1

            r_forest = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0, )
            scores = cross_validation.cross_val_score(r_forest, self.iss_features, self.labels, cv=k_folds,n_jobs=5)
            r_forest_full = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0)
            r_forest_full.fit(self.iss_features,self.labels)
            self.treedata[i,0] = n_trees
            self.treedata[i,1] = scores.mean()
            self.treedata[i,2] = scores.std()
            # now add the test dataset - score
            self.treedata[i,3] = r_forest_full.score(self.iss_validation_features, self.validation_labels)

            r_forest_lda = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0)
            r_forest_lda_full = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0)
            r_forest_lda_full.fit(self.lda_iss_features,self.labels)
            lda_scores = cross_validation.cross_val_score(r_forest_lda, self.lda_iss_features, self.labels, cv=k_folds,n_jobs=5)
            self.treedata[i,4] = lda_scores.mean()
            self.treedata[i,5] = lda_scores.std()
            self.treedata[i,6] = r_forest_lda_full.score(self.lda_iss_validation_features, self.validation_labels)


            r_forest_pca = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0)
            r_forest_pca_full = RandomForestClassifier(n_estimators=n_trees, n_jobs=5, max_depth=None, min_samples_split=1, random_state=0)
            r_forest_pca_full.fit(self.pca_iss_features,self.labels)
            pca_scores = cross_validation.cross_val_score(r_forest_pca, self.pca_iss_features, self.labels, cv=k_folds,n_jobs=5)
            self.treedata[i,7] = pca_scores.mean()
            self.treedata[i,8] = pca_scores.std()
            self.treedata[i,9] = r_forest_pca_full.score(self.pca_iss_validation_features, self.validation_labels)

    def pca(self,n_components = 6):
        self.pca = PCA(n_components)
        self.pca_iss_features = self.pca.fit_transform(self.iss_features)
        self.pca_iss_validation_features = self.pca.transform(self.iss_validation_features)

    def lda(self,n_components = 2, pca_reg = True, reg_dimensions = 10):
        self.lda = LinearDiscriminantAnalysis(n_components = n_components, solver='eigen', shrinkage='auto')
        #self.lda = LDA(n_components)
        if pca_reg:
            self.pca_reg = PCA(reg_dimensions)
            pca_reg_features = self.pca_reg.fit_transform(self.iss_features)
            self.lda_iss_features = self.lda.fit_transform(pca_reg_features,self.labels)
            pca_reg_validation_features = self.pca_reg.transform(self.iss_validation_features)
            self.lda_iss_validation_features = self.lda.transform(pca_reg_validation_features)
        else:
            self.lda_iss_features = self.lda.fit_transform(self.iss_features,self.labels)
            self.lda_iss_validation_features = self.lda.transform(self.iss_validation_features)

    def lda_run(self, k_folds = 5):
        self.r_forest_lda = RandomForestClassifier(n_estimators=2000,
                                                   n_jobs=5,
                                                   max_depth=None,
                                                   min_samples_split=2,
                                                   random_state =7,
                                                   max_leaf_nodes=None,
                                                   min_samples_leaf=2,
                                                   criterion='gini',
                                                   max_features='sqrt',
                                                   class_weight='balanced')

        self.lda_scores = cross_validation.cross_val_score(self.r_forest_lda,
                                                           self.lda_iss_features,
                                                           self.labels,
                                                           cv=k_folds,
                                                           n_jobs=5)
        print("Cross validation Random Forest performance LDA: Accuracy: %0.2f (std %0.2f)" % (self.lda_scores.mean()*100, self.lda_scores.std()*100))
        self.r_forest_lda.fit(self.lda_iss_features,self.labels)
        print(str(self.r_forest_lda.score(self.lda_iss_validation_features,
                                      self.validation_labels)*100)+ 'LDA test-set performance')

        y_true = self.validation_labels
        y_pred = self.r_forest_lda.predict(self.lda_iss_validation_features)
        target_names = ['S1','S2','S3','S4']
        report = classification_report(y_true, y_pred, target_names=target_names)
        print('Random forest report lda')
        print(report)

        ##### Hacky way to export features, so can optimise RF etc ######
        train_X = pd.DataFrame(self.lda_iss_features)
        train_y  = pd.DataFrame(self.labels)
        training = pd.concat([train_X,train_y], axis = 1)
        training.to_csv('/Users/Jonathan/Dropbox/Data_sharing_VMJC/training_lda.csv', index = False)

        test_X = pd.DataFrame(self.lda_iss_validation_features)
        test_y = pd.DataFrame(self.validation_labels)
        test = pd.concat([test_X,test_y], axis = 1)
        test.to_csv('/Users/Jonathan/Dropbox/Data_sharing_VMJC/test_lda.csv', index = False)

        train_X = pd.DataFrame(self.iss_features)
        train_y  = pd.DataFrame(self.labels)
        training = pd.concat([train_X,train_y], axis = 1)
        training.to_csv('/Users/Jonathan/Dropbox/Data_sharing_VMJC/training.csv', index = False)

        test_X = pd.DataFrame(self.iss_validation_features)
        test_y = pd.DataFrame(self.validation_labels)
        test = pd.concat([test_X,test_y], axis = 1)
        test.to_csv('/Users/Jonathan/Dropbox/Data_sharing_VMJC/test.csv', index = False)

    def pca_run(self,k_folds = 5):
        self.r_forest_pca = RandomForestClassifier(n_estimators=2000,n_jobs=5, max_depth=None, min_samples_split=1, random_state =0)
        self.pca_scores = cross_validation.cross_val_score(self.r_forest_pca, self.pca_iss_features, self.labels, cv=k_folds,n_jobs=5)
        print("Cross validation RF performance PCA: Accuracy: %0.2f (std %0.2f)" % (self.pca_scores.mean()*100, self.pca_scores.std()*100))

        self.r_forest_pca.fit(self.pca_iss_features,self.labels)
        print(str(self.r_forest_pca.score(self.pca_iss_validation_features, self.validation_labels))+ 'PCA test-set performance ')

    def run(self):

        r_forest = RandomForestClassifier(n_estimators=2000,n_jobs=5, max_depth=None, min_samples_split=1, random_state =0, class_weight='balanced')
        self._cross_validation(r_forest)
        print("Cross validation RF performance: Accuracy: %0.2f (std %0.2f)" % (self.scores.mean()*100, self.scores.std()*100))

        self.r_forest = RandomForestClassifier(n_estimators=2000,n_jobs=5, max_depth=None, min_samples_split=1, random_state=0, class_weight='balanced')
        self.r_forest.fit(self.iss_features,self.labels)

        print(str(self.r_forest.score(self.iss_validation_features, self.validation_labels))+ 'randomforest test-set performance')

        y_true = self.validation_labels
        y_pred = self.r_forest.predict(self.iss_validation_features)
        target_names = ['inter-ictal', 'ictal']
        target_names = ['S1','S2','S3','S4']
        t = classification_report(y_true, y_pred, target_names=target_names)
        print('Random forest report:')
        print(t)


        return None




'''

Make this into a class!
CHANGE TO POMEMGRANATE - CHECK WORKS ON PIP
'''
import pomegranate
from sklearn.preprocessing import normalize
import numpy as np

def make_hmm_model(emission_mat, transition_probs):
    model = pomegranate.HiddenMarkovModel('ndf')

    ictal_emissions    = {i:emission_mat[1,i] for i in range(emission_mat.shape[1])}
    baseline_emissions = {i:emission_mat[0,i] for i in range(emission_mat.shape[1])}

    ictal    = pomegranate.State(pomegranate.DiscreteDistribution(ictal_emissions   ), name = '1')
    baseline = pomegranate.State(pomegranate.DiscreteDistribution(baseline_emissions), name = '0')

    model.add_state(ictal)
    model.add_state(baseline)

    model.add_transition( model.start, ictal, 0.05 )
    model.add_transition( model.start, baseline, 99.95)

    model.add_transition( baseline, baseline, transition_probs[0,0] )
    model.add_transition( baseline, ictal,    transition_probs[0,1]  )
    model.add_transition( ictal, ictal   ,    transition_probs[1,1] )
    model.add_transition( ictal, baseline,    transition_probs[1,0]  )

    model.bake(verbose=False )
    return model

def get_state_emission_probs(emissions, annotated_states):
    n_states = len(np.unique(annotated_states))
    n_emiss  = len(np.unique(emissions))
    emis_mat = np.zeros(shape= (n_states,n_emiss))

    for i,label in enumerate(annotated_states.astype('int')):
        emis = emissions.astype('int')[i]
        emis_mat[label, emis] += 1
        emis_probs = normalize(emis_mat, axis = 1, norm='l1')
    return emis_probs

def get_state_transition_probs(labels):
    if len (labels.shape) > 1:
            labels = np.ravel(labels)

    tp = np.zeros(shape= (2,2))
    for i, label in enumerate(labels[:-1]):
        next_label = int(labels[i+1])
        label = int(label)
        tp[label,next_label] += 1

    tp = normalize(tp, axis = 1, norm='l1')
    return tp



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import ndf
from ndf.h5loader import H5File
from ndf.datahandler import DataHandler, NdfFile
from ndf.classifier import Classifier

try:
    import loading_subwindow, convert_ndf_window, library_subwindow, add_pred_features_subwindow, clf_subwindow
except:
    from . import loading_subwindow, convert_ndf_window, library_subwindow, add_pred_features_subwindow, clf_subwindow



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'predictions_loading_subwindow.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(582, 418)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(310, 130, 169, 23))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayoutWidget = QtWidgets.QWidget(Dialog)
        self.formLayoutWidget.setGeometry(QtCore.QRect(70, 40, 421, 71))
        self.formLayoutWidget.setObjectName("formLayoutWidget")
        self.formLayout_2 = QtWidgets.QFormLayout(self.formLayoutWidget)
        self.formLayout_2.setContentsMargins(0, 0, 0, 0)
        self.formLayout_2.setObjectName("formLayout_2")
        self.prediction_display = QtWidgets.QLineEdit(self.formLayoutWidget)
        self.prediction_display.setObjectName("prediction_display")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.prediction_display)
        self.h5_display = QtWidgets.QLineEdit(self.formLayoutWidget)
        self.h5_display.setObjectName("h5_display")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.h5_display)
        self.set_prediction_file = QtWidgets.QPushButton(self.formLayoutWidget)
        self.set_prediction_file.setMaximumSize(QtCore.QSize(170, 16777215))
        self.set_prediction_file.setObjectName("set_prediction_file")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.set_prediction_file)
        self.set_h5_folder = QtWidgets.QPushButton(self.formLayoutWidget)
        self.set_h5_folder.setMaximumSize(QtCore.QSize(141, 32))
        self.set_h5_folder.setObjectName("set_h5_folder")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.set_h5_folder)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.set_prediction_file.setText(_translate("Dialog", "Choose Prediction file"))
        self.set_h5_folder.setText(_translate("Dialog", "Choose h5 folder"))




def filterArray(array,window_size = 51,order=3):
    '''
    Simple for-loop based indexing for savitzky_golay filter

    Inputs:
    array:
    	array.shape[1] are datapoints - each row a trace, columns the datapoints
    	array should be <= 3d
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.


    '''

    import copy
    fData = copy.copy(array)
    ndimensions = len(fData.shape) # number of dimension
    if ndimensions == 1:
        fData = savitzky_golay(array,window_size,order)
    elif ndimensions == 2:
         for trace_i in range(array.shape[0]):
                    fData[trace_i,:] = savitzky_golay(array[trace_i,:],window_size,order)
    elif ndimensions == 3:
        for index in range(array.shape[2]):
                for index2 in range(array.shape[1]):
                    fData[:,index2,index] = savitzky_golay(array[:,index2,index],window_size,order)
    #else:
        #print "Jonny only bothered too (badly) code up to 3 array dimensions!"

    return fData

def savitzky_golay(y, window_size, order, deriv=0, rate=1):
    r"""Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688

       http://wiki.scipy.org/Cookbook/SavitzkyGolay
    """
    import numpy as np
    from math import factorial

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve( m[::-1], y, mode='valid')


import sys
import bokeh.plotting as bp
if sys.version_info > (3,):
    from bokeh.io import output_notebook
else:
    from bokeh.plotting import output_notebook
from .utils import filterArray
import numpy as np
import matplotlib.pyplot as plt

def basic_plot(data_dict, time_tuple = (0,3600), arg_dict = {'color' :'k', 'linewidth':'1.5'}):
    """
    Inputs:
        - data dictionary (ndf[id] returns this) containing 'time' and 'data' keys,
        - time tuple in seconds, default (0,3600). Determines time period to plot
        - arg_dict: dictionary of parameters to pass to the plotting function
    """
    data = data_dict['data']
    time_ = data_dict['time']

    fig = plt.figure(figsize = (15, 4))
    ax = fig.add_subplot(111)
    #ax = fig.add_axes()
    indexes = np.logical_and(time_>time_tuple[0], time_<time_tuple[1])
    ax.plot(time_[indexes],data[indexes], **arg_dict)
    ax.set_xlim(time_tuple[0], time_tuple[1])
    return ax


def plot(data_dict, nsec = 3600):
    if sys.version_info < (3,):
        print('Plotting function is not available for python 2.7, please upgrade to 3!')
        sys.exit()
    print('WARNING: This plotting downsamples the data too much! ')
    data = data_dict['data']
    time = data_dict['time']

    x_downsamp = np.linspace(0,nsec,num = 50*nsec)
    y_downsamp = np.interp(x_downsamp, time, data)

    filtered = filterArray(y_downsamp, window_size=51, order = 3)


    output_notebook()

    p = bp.figure(webgl=True,
              plot_width=900,
              plot_height=500,)

    p.grid.grid_line_alpha = 0
    p.xaxis.axis_label = 'Time (seconds)'
    p.yaxis.axis_label = 'Voltage'

    p.circle(x_downsamp, y_downsamp, size=4, legend='50hz interpolation',
          color='darkgrey', alpha=0.2)

    window_size = 30
    window = np.ones(window_size)/float(window_size)

    #data_avg = np.convolve(y_downsamp, window, 'valid')

    p.line(x_downsamp, filtered, legend='filtered', color='black')

    p.legend.location = "top_left"
    bp.show(p)



'''

This module is not currently being used

'''
from __future__ import print_function
import pickle
import os
import matplotlib.pyplot as plt
import numpy as np

import matplotlib.font_manager as fm
dir = os.path.dirname(__file__)
filename = os.path.join(dir, '../HelveticaNeue-Light.otf')
prop = fm.FontProperties(fname=filename)

def plot_traces(to_plot,
                start = 0,
                labels = None,
                savepath = None,
                filename = None,
                format_string = ".pdf",
                prob_thresholds = None, verbose = False):
    '''

    Args:
        to_plot: expecting this with
        start:
        labels: if None, traces are all assigned with 1's
        savestring:
        format_string:
        prob_thresholds: if not None, used to color code the index (0-sure- "k", 1-unsure "r")

    Returns:

    '''

    if labels is None:
        labels = np.ones(to_plot.shape[0])
    if prob_thresholds is None:
        prob_thresholds = np.zeros(to_plot.shape[0]).astype(int)

    colors = ['b','r','g','k']
    colors = ['k','r','g','b','purple']

    print('Plotting traces...')
    for section in range(int(np.ceil(to_plot.shape[0]/40.0))):


        fig = plt.figure(figsize=(8.27, 11.69), dpi=20)
        plt.axis('off')
        plt.title('Traces '+ str((section+start)*40)+ ':' + str(((section+start)+1)*40)+'  1:black 2:red 3:green 4:blue', fontproperties=prop, fontsize = 14)
        time = np.linspace(1,10,to_plot.shape[1])
        annotation_colors = ['k','r']

        if section == to_plot.shape[0]/40:
            n_plots = to_plot.shape[0]%40
            if verbose:
                print(str(section*40)+ ' : ' + str(((section)*40)+n_plots))
        else:
            n_plots = 40
            if verbose:
                print(str(section*40)+ ' : ' + str((section+1)*40)+',')

        for i in [ii + (section)*40 for ii in range(n_plots)]:
            ax = fig.add_subplot(20,2,(i%40)+1)
            ax.plot(time,to_plot[i,:], color = colors[int(labels[i])-1], linewidth = 0.5)
            ax.annotate(str(i+start), xy = (0,0.3), fontsize = 10, color = annotation_colors[prob_thresholds[i]], fontproperties=prop)

            ax.axis('off')
            ax.set_xlim((0,10))

            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.get_yaxis().tick_left()
            ax.get_xaxis().tick_bottom()

        if savepath:
            plt.savefig(os.path.join(savepath,filename+'_'+str(section+ (start+1)/40)+format_string))
            #plt.savefig(os.path.join(savepath,filename+'_'+str(section)+format_string))

        else:
            plt.show()

        plt.close('all')


def plot_traces_hdf5(to_plot,
                labels = None,
                savepath = None,
                filename = None,
                format_string = ".pdf",
                prob_thresholds = None,
                trace_len_sec = 5,
                verbose = False):

    if not format_string.startswith('.'):
        format_string = '.'+format_string

    if labels is None:
        labels = np.ones(to_plot.shape[0])
    if prob_thresholds is None:
        prob_thresholds = np.zeros(to_plot.shape[0]).astype(int)

    colors = ['k','r','g','b','purple']
    print('Plotting traces...')
    for section in range(int(np.ceil(to_plot.shape[0]/40.0))):

        fig = plt.figure(figsize=(8.27, 11.69), dpi=20)
        plt.axis('off')
        plt.title('Seconds '+ str(section*40*trace_len_sec)+ ':' + str((section+1)*40*trace_len_sec), fontsize = 14,fontproperties=prop)
        time = np.linspace(1,10,to_plot.shape[1])
        annotation_colors = ['k','r']

        if section == to_plot.shape[0]/40:
            n_plots = to_plot.shape[0]%40
            if verbose:
                print(str(section*40)+ ' : ' + str(((section)*40)+n_plots))
        else:
            n_plots = 40
            if verbose:
                print(str(section*40)+ ' : ' + str((section+1)*40)+',')

        for i in [ii + (section)*40 for ii in range(n_plots)]:
            ax = fig.add_subplot(20,2,(i%40)+1)
            ax.annotate(str(i), xy = (0,0.5), fontsize = 10,color = 'black', fontproperties=prop)

            ax.plot(time, to_plot[i,:], color = colors[int(labels[i])], linewidth = 0.5)
            ax.set_title(str(i*trace_len_sec), fontsize = 8, fontproperties=prop)
            ax.axis('off')
            ax.set_xlim((0,10))

            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.get_yaxis().tick_left()
            ax.get_xaxis().tick_bottom()

        if savepath:
            plt.savefig(os.path.join(savepath,filename+'_'+str(section)+format_string))

        plt.close('all')
    print('Done')



if __name__ == "__main__":
    training_tuple = pickle.load(open('../training_label_traces_tuple','rb'))
    training_tuple = pickle.load(open('../validation_label_traces_tuple','rb'))

    labels = training_tuple[0]
    data = training_tuple[1]
    print('plotting '+ str(data.shape[0])+ 'traces')
    plot_traces(data,labels,savestring='../validation ',format_string='.pdf')


from __future__ import print_function
import pickle

import numpy as np
import pandas as pd

import utils
from pyecog.light_code.network_loader import SeizureData
from pyecog.light_code.extractor import FeatureExtractor
from pyecog.light_code.make_pdfs import plot_traces


class Predictor():

    '''
    Class to predict state from unlabeled data.

    Data should be in th same format as the training data - stacked into traces and saved as a hdf5 file.
    '''

    def __init__(self, clf_pickle_path=None, threshold = 50):

        '''

        This currently (messily) using a fs_dict, for sampling frequency of files.
        And a tuple of skipfiles
        Args:
            clf_pickle_path: the pickled classifier to use
            fs_dict_path: dictionary made by abf_loader

        Returns:

        '''


        self.sg_filter_window_size = 7
        self.sg_filter_window_order = 3
        self.threshold = threshold # 'sureity' threshold

        self.classifier = pickle.load(open(clf_pickle_path,'rb'))
        self.lda = self.classifier.lda
        self.r_forest_lda = self.classifier.r_forest_lda


    def load_traces_to_classify(self,path, fs_dict_path='../pickled_fs_dictionary',):

        self.skipfiles = ('EX150515T11',
                        'EX180315T14',
                        'EX180515T4',
                        'EX200515T4.',)

        self.skip_dir = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/PV_ChR2/'


        self.fs_dict  = pickle.load(open(fs_dict_path,'rb'))

        # N.B only predicting with the lda r_forest at the moment
        #self.r_forest = self.classifier.r_forest

        self.dataobj = SeizureData(path, fs_dict = self.fs_dict)
        self.dataobj.load_data()
        f = open('../'+savestring+'_saved','wb')
        pickle.dump(self.dataobj,f)

    def assess_states(self, saved_path,
                      savestring = 'example',
                      pdf_savepath = '../',
                      make_pdfs = True):

        self.savestring = savestring
        self.pdf_savepath = pdf_savepath


        self.dataobj = pickle.load(open(saved_path,'rb'))

        self.norm_data = utils.normalise(self.dataobj.data_array)
        self.norm_data = utils.filterArray(self.norm_data,
                                           window_size= self.sg_filter_window_size,
                                           order=self.sg_filter_window_order)

        feature_obj = FeatureExtractor(self.norm_data)

        i_features = self.classifier.imputer.transform(feature_obj.feature_array)
        iss_features = self.classifier.std_scaler.transform(i_features)
        lda_iss_features = self.lda.transform(iss_features)

        # predict probability and also the actual state
        self.pred_table = self.r_forest_lda.predict_proba(lda_iss_features)*100
        self.preds = self.r_forest_lda.predict(lda_iss_features)

        # Make stuff for the excel sheet
        self.predslist = list(self.preds) # why need this?
        self.predslist[self.predslist == 4] = 'Baseline'
        self.max_preds = np.max(self.pred_table, axis = 1)
        self.threshold_for_mixed = np.where(self.max_preds < int(self.threshold),1,0) # 1 when below

        # do the 1st vs 2nd most likely states
        self.sorted_pred = np.sort(self.pred_table, axis = 1)
        self.ratio = np.divide(self.sorted_pred[:,2],self.sorted_pred[:,3])
        self.threshold_for_ratio = np.where(self.ratio > 0.5,1,0) # 1 when below

        # combine the two measures
        self.combined_pass = np.logical_or(self.threshold_for_mixed,self.threshold_for_ratio)

        self._string_fun2()
        self._write_to_excel()
        if make_pdfs:
            plot_traces(self.norm_data,
                        self.preds,
                        savepath= self.pdf_savepath + self.savestring,
                        #savestring = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pdfs0302/'+self.savestring,
                        prob_thresholds= self.combined_pass)

    def _string_fun2(self):
        '''
        This method is for the full data, vm gave in 2016/01
        '''
        self.nameframe =  pd.DataFrame(columns = ['Date', 'ID', 'File Start', 'File End', 'Pulse Time'])
        for i,f in enumerate(self.dataobj.filename_list):

            f =  f.split('/')[-1]
            try:
                date = f.split('X')[1].split('T')[0]
            except IndexError:
                date = f.split('x')[1].split('T')[0]
            t_start = '0'
            t_end   = f.split('_')[-1]
            t_onset = f.split('_')[1]
            transmitter = f.split('_')[0].split(date)[-1]
            #print [date, transmitter, t_start, t_end, t_onset]
            self.nameframe.loc[i] = [date, transmitter, t_start, t_end, t_onset]
            #print f

    def _string_fun(self):
        '''
        This is for the original style of file input...
        '''
        self.nameframe =  pd.DataFrame(columns = ['Date', 'ID', 'File Start', 'File End', 'Pulse Time'])

        for i,f in enumerate(self.dataobj.filename_list):
            #print f

            f =  f.split('/')[-1]
            s_brackets = f.split('[', 1)[1].split(']')[0]
            s_brackets_start = s_brackets.split('-')[0]
            t_end = s_brackets.split('-')[-1].split('s')[0]
            try:
                t_start = int(s_brackets_start)
            except:
                t_start = int(s_brackets_start.split(' ')[-1])
            date = f.split('X')[1].split('t')[0].strip('-')
            transmitter = f.split('_')[0].split('r')[-1]
            t_onset = float(f.split(']')[-1]) + t_start

            self.nameframe.loc[i] = [date, transmitter, t_start, t_end, t_onset]
        #print self.nameframe.head()

    def _write_to_excel(self):
        sheet = pd.DataFrame(self.pred_table)
        sheet.columns = ['State1','State2','State3','Baseline']

        pred = pd.DataFrame(self.predslist,columns=['Index'])

        max_preds = pd.DataFrame(self.max_preds)
        max_preds.columns = ['Max']

        master_exc =  pd.DataFrame(self.combined_pass,columns=['Exclude'])

        ratio = pd.DataFrame(self.ratio, columns = ['2v1_Ratio'])

        frames = [self.nameframe, sheet, max_preds, ratio, master_exc, pred]
        vmsheet = pd.concat(frames,axis = 1)
        #print vmsheet.head()
        writer = pd.ExcelWriter('/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/'+self.savestring+'.xlsx',engine = 'xlsxwriter')
        vmsheet.to_excel(writer,index = True,sheet_name = 'Pulse prediction' )
        workbook = writer.book
        worksheet = writer.sheets['Pulse prediction']
        percent_fmt = workbook.add_format({'num_format': '0.00', 'bold': False})
        format1 = workbook.add_format({'bg_color': '#FFC7CE',
                               'font_color': '#9C0006'})
        worksheet.set_column('G:J',12,percent_fmt)

        color_range = "K2:K{}".format(len(self.dataobj.filename_list)+1)
        worksheet.conditional_format(color_range, {'type': 'cell',
                                                   'criteria': '<=',
                                           'value': self.threshold,
                                           'format': format1})

        ratio_range = "L2:L{}".format(len(self.dataobj.filename_list)+1)
        worksheet.conditional_format(ratio_range, {'type': 'cell',
                                                   'criteria': '>=',
                                           'value': 0.5,
                                           'format': format1})
        ratio_range = "M2:M{}".format(len(self.dataobj.filename_list)+1)
        worksheet.conditional_format(ratio_range, {'type': 'cell',
                                                   'criteria': '=',
                                           'value': 1,
                                           'format': format1})
        writer.save()

def main():
    # there is also a pickled classifier in the directory just above here
    x = Predictor( clf_pickle_path = '/Volumes/LACIE SHARE/VM_data/classifier_hdf5data/pickled_classifier_20160223',
                   threshold= 50)


    makepdfs = True
    x.assess_states(savestring='PV_ARCH_predictions_20160310',
                    pdf_savepath='/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pdfs0310/',
                    saved_path = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pickled_tensec_dobj/PV_ARCH_pickled_tensec',
                    make_pdfs= makepdfs)

    '''
    raw_path = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/PV_ARCH/ConvertedFiles/',
    x.assess_states(raw_path = None,
                    savestring='PV_CHR2_predictions_20160310',
                    raw_load = False,
                    pdf_savepath='/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pdfs0310/',
                    saved_path = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pickled_tensec_dobj/PV_CHR2_pickled_tensec',
                    make_pdfs= makepdfs)

    x.assess_states(raw_path = None,
                    savestring='SOM_CHR2_predictions_20160310',
                    raw_load = False,
                    pdf_savepath='/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pdfs0310/',
                    saved_path = '/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/pickled_tensec_dobj/SOM_CHR2_pickled_tensec',
                    make_pdfs= makepdfs)
    '''
if __name__ == "__main__":
    main()





    '''
    ###################################
    # adding to the training dataset!
    for_training = True
    certified_training = []
    if for_training:
        #print 'here'
        t_labels = pd.read_excel('/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/labelling_for_training/090216_PVArch_test0_7.xlsx').values
        #print t_labels
        for i in range(t_labels.shape[0]):
            if t_labels[i,1] != 4: # in this excel doc we usde 4 for mixed state! :/ and 0 for baseline
                certified_training.append(i)


        #print t_labels[certified_training,:]
        f = open('/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/labelling_for_training/new_training_data_2016_02_09','wb')
        pickle.dump(self.dataobj.data_array[certified_training,:],f)
        print self.dataobj.data_array[certified_training,:].shape

        t_labels_vet = t_labels[certified_training,1]
        t_labels_vet[t_labels_vet==0] = 4 # need to correct to use the baseline index 4, as normal
        f = open('/Volumes/LACIE SHARE/VM_data/All_Data_Jan_2016/labelling_for_training/new_training_labels_2016_02_09','wb')
        pickle.dump(t_labels_vet,f)
        print t_labels_vet.shape
        print t_labels_vet

    exit()

    ###################################
    '''


import glob, os, numpy, sys
try:
    import stfio
except:
    sys.path.append('C:\Python27\Lib\site-packages')
    import stfio

def main():

    
    
    searchpath = os.getcwd()
    exportdirectory = searchpath+'/ConvertedFiles/'

    # Make export directory
    if not os.path.exists(exportdirectory):
        os.makedirs(exportdirectory)
        
    # Walk through and find abf files   

    pattern = '*.abf'
    datafilenames = glob.glob(pattern)
    if datafilenames:
        for filename in datafilenames:
            print ('Converting '+str(filename))
            data = stfio.read(filename,ftype = "abf")
            x = data.aspandas()
            x = x.values
            numpy.save(exportdirectory+filename[0:-4],x)

           

if __name__ == '__main__':
  main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'add_prediction_features.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_make_features(object):
    def setupUi(self, make_features):
        make_features.setObjectName("make_features")
        make_features.resize(755, 562)
        self.formLayoutWidget = QtWidgets.QWidget(make_features)
        self.formLayoutWidget.setGeometry(QtCore.QRect(70, 70, 537, 131))
        self.formLayoutWidget.setObjectName("formLayoutWidget")
        self.formLayout_2 = QtWidgets.QFormLayout(self.formLayoutWidget)
        self.formLayout_2.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldsStayAtSizeHint)
        self.formLayout_2.setContentsMargins(0, 0, 0, 0)
        self.formLayout_2.setObjectName("formLayout_2")
        self.set_h5_folder = QtWidgets.QPushButton(self.formLayoutWidget)
        self.set_h5_folder.setMinimumSize(QtCore.QSize(200, 0))
        self.set_h5_folder.setMaximumSize(QtCore.QSize(200, 32))
        self.set_h5_folder.setObjectName("set_h5_folder")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.set_h5_folder)
        self.h5_display = QtWidgets.QLineEdit(self.formLayoutWidget)
        self.h5_display.setMinimumSize(QtCore.QSize(300, 0))
        self.h5_display.setObjectName("h5_display")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.h5_display)
        self.label = QtWidgets.QLabel(self.formLayoutWidget)
        self.label.setObjectName("label")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.chunk_len_box = QtWidgets.QLineEdit(self.formLayoutWidget)
        self.chunk_len_box.setObjectName("chunk_len_box")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.chunk_len_box)
        self.label_3 = QtWidgets.QLabel(self.formLayoutWidget)
        self.label_3.setObjectName("label_3")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.cores_to_use = QtWidgets.QLineEdit(self.formLayoutWidget)
        self.cores_to_use.setObjectName("cores_to_use")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.cores_to_use)
        self.run_peakdet_checkBox = QtWidgets.QCheckBox(self.formLayoutWidget)
        self.run_peakdet_checkBox.setText("")
        self.run_peakdet_checkBox.setObjectName("run_peakdet_checkBox")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.run_peakdet_checkBox)
        self.label_2 = QtWidgets.QLabel(self.formLayoutWidget)
        self.label_2.setObjectName("label_2")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.progressBar = QtWidgets.QProgressBar(make_features)
        self.progressBar.setGeometry(QtCore.QRect(60, 342, 541, 31))
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.progress_bar_label = QtWidgets.QLabel(make_features)
        self.progress_bar_label.setGeometry(QtCore.QRect(60, 330, 461, 16))
        self.progress_bar_label.setObjectName("progress_bar_label")
        self.extract_features_button = QtWidgets.QPushButton(make_features)
        self.extract_features_button.setGeometry(QtCore.QRect(280, 250, 181, 32))
        self.extract_features_button.setObjectName("extract_features_button")
        self.logpath_dsplay = QtWidgets.QLabel(make_features)
        self.logpath_dsplay.setGeometry(QtCore.QRect(60, 380, 600, 16))
        self.logpath_dsplay.setMinimumSize(QtCore.QSize(600, 0))
        self.logpath_dsplay.setObjectName("logpath_dsplay")
        self.hidden_label = QtWidgets.QLabel(make_features)
        self.hidden_label.setGeometry(QtCore.QRect(60, 300, 600, 16))
        self.hidden_label.setMinimumSize(QtCore.QSize(600, 0))
        self.hidden_label.setText("")
        self.hidden_label.setObjectName("hidden_label")
        self.label_4 = QtWidgets.QLabel(make_features)
        self.label_4.setGeometry(QtCore.QRect(180, 30, 441, 16))
        self.label_4.setObjectName("label_4")

        self.retranslateUi(make_features)
        QtCore.QMetaObject.connectSlotsByName(make_features)

    def retranslateUi(self, make_features):
        _translate = QtCore.QCoreApplication.translate
        make_features.setWindowTitle(_translate("make_features", "Dialog"))
        self.set_h5_folder.setText(_translate("make_features", "Choose h5 folder"))
        self.label.setText(_translate("make_features", "Timewindow length (s)"))
        self.chunk_len_box.setText(_translate("make_features", "5"))
        self.label_3.setText(_translate("make_features", "Number of CPU cores to use"))
        self.cores_to_use.setText(_translate("make_features", "all"))
        self.label_2.setText(_translate("make_features", "Use peakdet features?"))
        self.progress_bar_label.setText(_translate("make_features", "Progress:"))
        self.extract_features_button.setText(_translate("make_features", "Extract Features"))
        self.logpath_dsplay.setText(_translate("make_features", "Path to Logfile:"))
        self.label_4.setText(_translate("make_features", "** EVENTUALLY WILL IMPLEMENT EXTRACTION FROM NDFs **"))




import os
from os.path import join

import numpy as np

import utils


class LFPData(object):
    """
    Class for one section of data before the light pulse.
    """
    def __init__(self, path,preprocess, window_len, target_fs = 512, label = None, original_fs_dict = {}, filter = True, to_dwnsmple = False ):

        self.original_fs_dict = original_fs_dict
        self.target_fs = target_fs # this is not being used at the moment
        self.window_len = window_len*target_fs
        self.label = None
        self.filename = path

        print path
        self.alldata = np.load(path)
        print 'here'
        print self.alldata.shape
        #print self.alldata.shape
        if preprocess:

            key = self.filename.split('/')[-1].split('.')[0]
            print key
            try:
                self.original_fs = self.original_fs_dict[key]
            except KeyError:
                print 'Did not find key in fs dictonary!'
                # insert calculation code!

                if to_dwnsmple:
                    print 'got a shitty argument INCORRECT!'
                    self.original_fs =( to_dwnsmple * 500)/1000
                    print self.original_fs, 'is original rate in seconds'

                else:
                    exit(1)


            self.downsample_rate = int(self.original_fs * 1000/ float(self.target_fs)) # for converting between khz to 500 - in secconds


            # if more advanced can make a method for this
            #print "downsampled!"
            print(self.filename[-65:-3]+'...')
            self.fullfs = self.alldata[:,:]
            print self.alldata.shape, 'is now alldatashape'
            if filter:
                print 'filtering!'
                self.alldata[:,1] = utils.filterArray(self.alldata[:,1])
                print self.alldata.shape
                print self.downsample_rate, 'is downsample_rate'
            self.alldata = self.alldata[::self.downsample_rate,:]
            file_len = str(self.alldata.shape[0]/500)
            print file_len


        self.lfpdata = self.alldata[:,1]


        if self.alldata.shape[1]>1:
            self.lightindexes = self._getLightIndexes(self.fullfs[:,0])
            self.n_lightpulses = len(self.lightindexes)

            self.data_array = []
            self.label_array = []
            self.name_array = []
            #print len(self.lightindexes)
            for lightindextuple in self.lightindexes:
                #print lightindextuple
                i = lightindextuple[0] - (self.window_len) # in case we
                if i > 0:
                    data_section = self.lfpdata[i:lightindextuple[0]]
                    self.data_array.append(data_section)
                    self.label_array.append(label)
                    self.name_array.append(self.filename[:-4]+'_'+str(lightindextuple[0]/float(self.target_fs))+'_outof_'+file_len)

            self.data_array = np.array(self.data_array)
            #self.data_array -= np.mean(self.data_array,axis = 0)
            # remove glitches
            #print 'remove glitches in lfpdata.py ran '
            print '*******', self.data_array.shape, 'is data array shape ********'
            self.threshold = np.std(self.data_array, axis = 1)*7
            self.mask = np.where(self.data_array>self.threshold[:,None],0,1)
            self.mask2 = np.where(self.data_array<-self.threshold[:,None],0,1)
            self.mask += self.mask2
            self.mask /= 2
            self.data_array *= self.mask
            ### Finish removing glitches ###
            self.label_array = np.array(self.label_array)
            self.name_array = np.array(self.name_array)
            #plt.figure(figsize=(12,6))
            #plt.plot(self.data_array.T)
            #plt.title(self.filename)
            assert len(self.data_array.shape) ==2



    def _getLightIndexes(self,light, plot = False):

        threshold =  ((max(light)-min(light))/2)+min(light)
        #threshold = 1000
        print threshold, 'is threshold', self.original_fs
        mask = np.where(light>threshold,1,0)
        mask1 = np.roll(mask[:],1)
        mask = abs(np.subtract(mask,mask1))
        lightindexes = np.where(mask==1)[0]
        print lightindexes.shape
        if not lightindexes.shape[0]%2 == 0:
            raise AssertionError('The file has an odd number of light pulses,\
                                 cannot automatically detect the pulse times')
        light_index_list = []

        if plot:
            import matplotlib.pyplot as plt
        for i in range(0, lightindexes.shape[0], 2):
            length = lightindexes[i+1]-lightindexes[i]
            if length > 9.5*self.original_fs*1000 and length < 10.5*self.original_fs*1000:
                pulse = (lightindexes[i]/self.downsample_rate,lightindexes[i+1]/self.downsample_rate)
                #print (pulse[1]-pulse[0])/float(512), 'seconds after ds'
                light_index_list.append(pulse)
                color = 'k'
            else:
                color = 'r'
            if plot:
                print i, (lightindexes[i+1]-lightindexes[i])/float(self.original_fs*1000)
                plt.plot(light[lightindexes[i]-1*self.original_fs*1000:lightindexes[i+1]+1000*self.original_fs], color = color)
            #print light[lightindexes[i]-512], light[lightindexes[i]+512]
        if plot:
            plt.show()
        return light_index_list

class SeizureData(object):
    '''
    Class to load all files in a base directory
    '''
    def __init__(self, base_dir_path, fs_dict = {}, target_fs=512,to_dwnsmple = False):
        '''

        '''
        self.target_fs = target_fs
        self.to_dwnsmple = to_dwnsmple # this is bullshit for the training data from vincent

        self.original_fs_dict = fs_dict

        if not os.path.isdir(base_dir_path):
            raise ValueError('%s is not a directory.' % base_dir_path)
        self.base_dir = base_dir_path

    def load_data(self, n_states = 3, window = 10, preprocess = True,):
        """
        Loads data for a network state and type of data into class variable.
        No output.

        Inputs:


        """
        self.n_states = n_states # this is a throw back from when we had c1, c2, c3 folders
        self.window = window
        self.filenames = self._get_filenames() # own method in case we want to get clever
        self.data_array_list = []
        self.label_colarray_list = []
        self.filename_list = []
        print self.filenames

        for i, filename in enumerate(self.filenames):
            print filename
            #print float(i)/float(len(self.filenames))*100.," percent complete         \r",
            # Each call of _load_data_from_file appends data to features_train
            self.temp_data,self.temp_label, self.name = self._load_data_from_filename(filename, preprocess=preprocess,to_dwnsmple = self.to_dwnsmple)
            self.data_array_list.append(self.temp_data)
            self.label_colarray_list.append(self.temp_label[:])
            self.filename_list.append(self.name[:])
            #print self.temp_label
            #print self.temp_label.shape

        self.data_array = np.vstack((self.data_array_list))
        self.label_colarray = np.hstack((self.label_colarray_list))
        self.label_colarray = np.reshape(self.label_colarray, (self.label_colarray.shape[0],1))
        self.filename_list = np.hstack((self.filename_list))
        #print self.data_array.shape
        #print self.label_colarray.shape
        #print self.filename_list.shape
        #print "\nDone"

    def _get_filenames(self):

        filenames = [join(self.base_dir, name) for name in os.listdir(self.base_dir) if name[-4:] == '.npy']
        #print filenames

        if len(filenames) == 0:

            cfolders =  [os.path.join(self.base_dir, c_string) for c_string in os.listdir(self.base_dir) if c_string[0] == 'c']

            filenames = []
            for folder in cfolders:
                for name in os.listdir(folder):
                    if name[-4:] == '.npy':
                        filenames.append(os.path.join(folder, name))

        return filenames

    def _load_data_from_filename(self,filename, preprocess = True,to_dwnsmple = False ):

        for i in range(self.n_states):
            if filename.find('c'+str(i+1)) >-1: # in case pre-classified in folders
                label = i+1
            else:
                label = 0
        lfp_data = LFPData(filename,preprocess, self.window, label = label, target_fs=self.target_fs, original_fs_dict=self.original_fs_dict,to_dwnsmple = to_dwnsmple )

        return lfp_data.data_array, lfp_data.label_array, lfp_data.name_array

    def extract_feature_array(self,feature_extractors = [None]):
        self.features = []
        self.feature_names = []
        for extractor in feature_extractors:
            self.features.append(extractor.extract(self.data_array))
            self.feature_names.append(extractor.names)
        self.feature_names = [item for sublist in self.feature_names for item in sublist] # flatten
        print self.feature_names, 'ARE THE FEATURES USED'
        self.features = np.hstack([self.features[i] for i in range(len(self.features))])


import sys
import struct
import os
import time
import logging

import pandas as pd
import h5py
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as ss
from scipy import signal, stats
#from line_profiler import LineProfiler

if sys.version_info < (3,):
    range = xrange
'''
# decorator needed when profiling
def lprofile():
    def inner(func):
        def profiled_func(*args, **kwargs):
            try:
                profiler = LineProfiler()
                profiler.add_function(func)

                profiler.enable_by_count()
                return func(*args, **kwargs)
            finally:
                profiler.print_stats()
        return profiled_func
    return inner
'''
class NdfFile:
    """
    TODO:
     - glitch detection is a little messy, relying on bindings.
     - bad messages doesnt delete if 2 messages in window for one (minor but would be better to delete both)
     - Find out if they ever start recording half way through - for auto fs detection
     - Code up printing the ndf object __repr__
     - Clean up unused  __init__ attributes

    Class to load ndf binary files.

    The NDF file starts with a header of at least twelve bytes:
        - The first four bytes spell the NDF identifier " ndf". The identifier is then
          followed by three four-byte big-endian numbers.
        - The first number is the address of the meta-data string, i.e. the byte offset from
          the first byte of the file (indexed from 0). This is therefore the number of bytes
          to skip from the start of the file to begin reading the meta-data.
        - The second number is the address of the first data byte. The data extends to the end of the file.
        - The third number is the actual length of the meta-data string, as it was last written.
          If this number is zero, any routines dealing with the meta-data string must determine
          the length of the string themselves.

    The messages in the data recorder's message buffer are each four bytes long. The bytes of each message are listed in
    the table below. The Channel Number is used to identify the source of the message. Channel number zero is reserved for
    clock messages. Following the channel number, each message contains a sixteen-bit data word. In the case of transmitters
    , the sixteen-bit data word is a digitized voltage. The last byte of the message is a timestamp.

    Byte	Contents
    0	Channel Number
    1	Most Significant Data Byte
    2	Least Significant Data Byte
    3	Timestamp or Version Number

    All messages with channel number zero are clock messages. This channel acts as a reference clock that
    is subsequently used to align the data messages from the transmitter channels and do error correction.
    The messages in this channel are generated at a frequency of 128 Hz.

    Each ndf file typically encodes 1 hour of data at 512 Hz, although it can also encode data at other frequencies
    (e.g. 1024 Hz) it does so for up to 14 transmitters. Each channel sends a message roughly 4 times for every
    channel 0 message (because they are operating at 512 Hz, while the clock is at 128 Hz).

    """

    def __init__(self, file_path, verbose = False, fs = 'auto', amp_factor = 200):

        self.filepath = file_path

        #  some unused
        self.tid_set = set()
        self.tid_to_fs_dict = {}
        self.tid_raw_data_time_dict  = {}
        self.tid_data_time_dict = {}
        self.resampled = False

        self.file_label = file_path.split('/')[-1].split('.')[0]
        self.identifier = None
        self.data_address = None
        self.metadata = None

        self.t_stamps = None
        self.read_ids = None
        self.fs = fs

        self._n_possible_glitches = None
        self._glitch_count        = None
        self._plot_each_glitch    = None
        self.read_id = None

        self.verbose = verbose

        self.file_time_len_sec = 3600
        self.micro_volt_div = 0.4 # this is the dac units?

        # firmware dependent:
        self.clock_tick_cycle = 7.8125e-3  # the "big" clock messages are 128Hz, 1/128 = 7.8125e-3
        self.clock_division = self.clock_tick_cycle / 256.0 # diff values from one byte

        self._read_file_metadata()
        self.get_valid_tids_and_fs()

    def __getitem__(self, item):
        #assert type(item) == int
        assert item in self.tid_set, 'ERROR: Invalid tid for file'
        return self.tid_data_time_dict[item]

    def _read_file_metadata(self):
        with open(self.filepath, 'rb') as f:

            f.seek(0)
            self.identifier = f.read(4)
            assert (self.identifier == b' ndf')

            meta_data_string_address = struct.unpack('>I', f.read(4))[0]
            self.data_address = struct.unpack('>I', f.read(4))[0]
            meta_data_length = struct.unpack('>I', f.read(4))[0]

            if meta_data_length != 0:
                f.seek(meta_data_string_address)
                self.metadata = f.read(meta_data_length)
                # need to handle the fact it is in bytes?
                #print ('\n'.join(self.metadata.split('\n')[1:-2]))
                #print (self.metadata)
            else:
                print('meta data length unknown - not bothering to work it out...')

    def get_valid_tids_and_fs(self):
        """
        - Here work out which t_ids are in the file and their
          sampling frequency. Threshold of at least 5000 datapoints!
        """
        f = open(self.filepath, 'rb')
        f.seek(self.data_address)
        self._e_bit_reads = np.fromfile(f, dtype = 'u1')
        self.transmitter_id_bytes = self._e_bit_reads[::4]
        tid_message_counts = pd.Series(self.transmitter_id_bytes).value_counts()
        possible_freqs = [256,512,1024]
        for tid, count in tid_message_counts.iteritems():
            if count > 5000 and tid != 0: # arbitrary threshold to exclude glitches
                error = [abs(3600 - count/fs) for fs in possible_freqs]
                if self.fs == 'auto':
                    self.tid_to_fs_dict[tid] = possible_freqs[np.argmin(error)]
                else:
                    self.fs = float(self.fs)
                    self.tid_to_fs_dict[tid] = self.fs
                self.tid_set.add(tid)
                self.tid_raw_data_time_dict[tid]  = {}
                self.tid_data_time_dict[tid] = {}
        logging.info(self.filepath)
        logging.info('Valid ids and freq are: '+str(self.tid_to_fs_dict))

    #@lprofile()
    def glitch_removal(self, plot_glitches=False, print_output=False,
                       plot_sub_glitches = False, tactic = 'mad'):
        """
        Tactics can either be 'std', 'mad','roll_med', 'big_guns'
        """
        for tid in self.read_ids:
            # create binding between tid data and the data to deglitch
            self.data_to_deglitch = self.tid_data_time_dict[tid]['data']
            self.time_to_deglitch = self.tid_data_time_dict[tid]['time']
            #print (self.data_to_deglitch is self.tid_data_time_dict[tid]['data'])
            self._n_possible_glitches = 0
            self._glitch_count        = 0
            self._plot_each_glitch = plot_sub_glitches

            if plot_glitches:
                plt.figure(figsize = (15, 4))
                plt.plot(self.time_to_deglitch , self.data_to_deglitch, 'k')
                plt.title('Full raw trace');plt.xlabel('Time (seconds)')
                plt.xlim(0,self.time_to_deglitch[-1])
                plt.show()

            if tactic == 'std':
                crossing_locations = self._stddev_based_outlier()
                self._check_glitch_candidates(crossing_locations)

            elif tactic == 'mad':
                crossing_locations = np.where(self._mad_based_outlier())[0]
                self._check_glitch_candidates(crossing_locations)

            elif tactic == 'roll_med':
                crossing_locations = np.where(self._rolling_median_based_outlier())[0]
                self._check_glitch_candidates(crossing_locations)

            elif tactic == 'big_guns':
                crossing_locations = np.where(self._rolling_median_based_outlier())[0]
                self._check_glitch_candidates(crossing_locations)
                crossing_locations = np.where(self._mad_based_outlier())[0]
                self._check_glitch_candidates(crossing_locations)
                crossing_locations = self._stddev_based_outlier()
                self._check_glitch_candidates(crossing_locations)
            else:
                print ('Please specify detection tactic: ("mad","roll_med","big_guns", "std")')
                raise

            logging.debug('Tid '+str(tid)+': removed '+str(self._glitch_count)+' datapoints as glitches. There were '+str(self._n_possible_glitches)+' possible glitches.')
            if self.verbose:
                print('Tid '+str(tid)+': removed '+str(self._glitch_count)+' datapoints as glitches. There were '+str(self._n_possible_glitches)+' possible glitches.')

            if plot_glitches:
                plt.figure(figsize = (15, 4))
                plt.plot(self.time_to_deglitch , self.data_to_deglitch, 'k')
                plt.title('De-glitched trace');plt.xlabel('Time (seconds)')
                plt.xlim(0,self.time_to_deglitch[-1] )
                plt.show()

            #self.tid_data_time_dict[tid]['data'] = self.data_to_deglitch[:]
            #self.tid_data_time_dict[tid]['time'] = self.time_to_deglitch[:]

    def _mad_based_outlier(self, thresh=3.5):
        """
        From stackoverflow?

        points : An numobservations by numdimensions array of observations
        thresh : The modified z-score to use as a threshold. Observations with
            a modified z-score (based on the median absolute deviation) greater
            than this value will be classified as outliers.
        """
        points = self.data_to_deglitch
        if len(points.shape) == 1:
            points = points[:,None]
        median = np.median(points, axis=0)
        diff = np.sum((points - median)**2, axis=-1)
        diff = np.sqrt(diff)
        med_abs_deviation = np.median(diff)
        modified_z_score = 0.6745 * diff / med_abs_deviation
        return modified_z_score > thresh

    def _rolling_median_based_outlier(self, threshold = 1):
        data = self.data_to_deglitch
        if len(data.shape) == 1:
            data = data[:, None]
        df = pd.DataFrame(data, columns=['raw'])
        df['rolling'] = df['raw'].rolling(window=10, center =
        True).median().fillna(method='bfill').fillna(method='ffill')
        difference = np.abs(df['raw'] - df['rolling'])
        #inlier_idx = difference < threshold
        outlier_idx = difference > threshold
        n_glitch = sum(abs(outlier_idx))
        if n_glitch > 200:
            logging.warning('Warning: more than 200 glitches detected! n_glitch = '+str(n_glitch))
        return outlier_idx


    def _stddev_based_outlier(self, x_std_threshold=10):
        std_dev = np.std(self.data_to_deglitch)
        mean_point = np.mean(self.data_to_deglitch)
        threshold = std_dev * x_std_threshold + mean_point
        crossing_locations = np.where(self.data_to_deglitch > threshold)[0]
        return crossing_locations

    def _check_glitch_candidates(self,crossing_locations, diff_threshold=10,):

        self._n_possible_glitches += len(crossing_locations)
        # check local difference is much bigger than the mean difference between points
        glitch_count = 0
        std_dev = np.std(self.data_to_deglitch)
        for location in crossing_locations:
            i = location - 1
            ii = location + 1
            try:
                if abs(self.data_to_deglitch[location] - self.data_to_deglitch[ii]) > diff_threshold * std_dev:
                    # plot glitches to be removed if plotting option is on
                    if self._plot_each_glitch:
                        plt.figure(figsize = (15, 4))
                        plt.plot(self.time_to_deglitch[location - 512:location + 512],
                                 self.data_to_deglitch[location - 512:location + 512], 'k')
                        plt.ylabel('Time (s)'); plt.title('Glitch '+str(glitch_count+1))
                        plt.show()
                    try: # todo work out what this does?
                        value = self.data_to_deglitch[i] + (self.time_to_deglitch[location] - self.time_to_deglitch[i]) * (
                        self.data_to_deglitch[ii] - self.data_to_deglitch[i]) / (self.data_to_deglitch[ii] - self.data_to_deglitch[i])
                        self.data_to_deglitch[location] = value
                    except IndexError:
                        pass
                    glitch_count += 1

            except IndexError:
                pass
        self._glitch_count += glitch_count


    def correct_sampling_frequency(self):
        '''
        Remeber, this is acting on the modified data (bad message and glitch already)
        so self.tid_data_time dict
        :return:
        '''
        # this occurs after bad messages, so working with data ditc
        # first check that we are not interpolating datapoints for more than 1 second?

        for tid in self.read_ids:
            max_interp = max(np.diff(self.tid_data_time_dict[tid]['time']))
            diff_array = np.diff(self.tid_data_time_dict[tid]['time'])
            second_gaps = diff_array[diff_array>1]
            seconds_missing = np.sum(second_gaps)
            n_second_gaps   = second_gaps.shape[0]
            if n_second_gaps >0:
                logging.debug('Tid '+str(tid)+': File contained '+ str(n_second_gaps)+' gaps of >1 second. Total Missing: '+str(seconds_missing)+ ' s.')
            try:
                assert max_interp < 2.0
            except:
                logging.warning(str(os.path.split(self.filepath)[1])+ ' Tid '+str(tid)+': You interpolated (at least once) for greater than two seconds! ('+ str('{first:.2f}'.format(first = max_interp))+' sec)')

            # do linear interpolation between the points, where !nan
            regularised_time = np.linspace(0, 3600.0, num= 3600 * self.tid_to_fs_dict[tid])
            not_nan = np.logical_not(np.isnan(self.tid_data_time_dict[tid]['data']))
            self.tid_data_time_dict[tid]['data'] = np.interp(regularised_time,
                                                             self.tid_data_time_dict[tid]['time'][not_nan],
                                                             self.tid_data_time_dict[tid]['data'][not_nan])

            self.tid_data_time_dict[tid]['time'] = regularised_time

            if self.verbose:
                print('Tid '+str(tid)+': regularised fs to '+str(self.tid_to_fs_dict[tid])+' Hz '+str(self.tid_data_time_dict[tid]['data'].shape[0]) +' datapoints')

        self._resampled = True


    def save(self, save_file_name = None):
        """
        Saves file in h5 format. Will only save the tid/tids that have loaded.
        Args:
            save_file_name:
        """
        if not save_file_name:
            hdf5_filename = self.filepath.strip('.ndf')+'_Tid_'+''.join(str([tid for tid in self.read_ids]))+ '.h5'
        else:
            hdf5_filename = save_file_name + '.h5'

        with h5py.File(hdf5_filename, 'w') as f:
            f.attrs['num_channels'] = len(self.read_ids)
            f.attrs['t_ids'] = list(self.read_ids)
            f.attrs['fs_dict'] = str(self.tid_to_fs_dict)
            file_group = f.create_group(os.path.split(self.filepath)[1][:-4])

            for tid in self.read_ids:
                transmitter_group = file_group.create_group(str(tid))
                transmitter_group.attrs['fs'] = self.tid_to_fs_dict[tid]
                transmitter_group.attrs['tid'] = tid
                transmitter_group.create_dataset('data',
                                                 data=self.tid_data_time_dict[tid]['data'],
                                                 compression = "gzip", dtype='f4',
                                                 chunks = self.tid_data_time_dict[tid]['data'].shape)
                # todo WHY ARE YOU SAVING TIME?!!!! Just generate it when you load up h5
                transmitter_group.create_dataset('time',
                                                 data=self.tid_data_time_dict[tid]['time'],
                                                 compression = "gzip", dtype='f4',
                                                 chunks = self.tid_data_time_dict[tid]['time'].shape)
                transmitter_group.attrs["resampled"] = self._resampled

            f.close()

            #print f.attrs['fs_dict']
        if self.verbose:
            print('Saved data as:'+str(hdf5_filename)+ ' Resampled = ' + str(self._resampled))

    def _merge_coarse_and_fine_clocks(self):
        # convert timestamps into correct time using clock id
        t_clock_data = np.zeros(self.voltage_messages.shape)
        t_clock_data[self.transmitter_id_bytes == 0] = 1 # this is big ticks
        corse_time_vector = np.cumsum(t_clock_data) * self.clock_tick_cycle
        fine_time_vector = self.t_stamps_256 * self.clock_division
        self.time_array = fine_time_vector + corse_time_vector

    def load(self, read_ids = [],
             auto_glitch_removal = True,
             auto_resampling = True,
             auto_filter = True,
             scale_to_mode_std = False):
        '''
        N.B. Should run glitch removal before high pass filtering and auto resampling... If unhappy with glitches,
        turn off filtering and the resampling and then run their methods etc.

        Args:
            read_ids: ids to load, can be integer of list of integers
            auto_glitch_removal: to automatically detect glitches with default tactic median abs deviation
            auto_resampling: to resample fs to regular sampling frequency
            auto_filter : high pass filter traces at default 1 hz
            scale_to_mode_std: high pass filter (default 1 hz) and scale to mode std dev 5 second blocks of trace
            WARNING: This is more for visualisation of what the feature extractor is working on. TO keep things
            simple, when saving HDF5 files, save non-scaled.

        Returns:
            data and time is stored in self.tid_data_time_dict attribute. Access data via obj[tid]['data'].


        '''

        self.read_ids = read_ids
        logging.info('Loading '+ self.filepath +'read ids are: '+str(self.read_ids))
        if read_ids == [] or str(read_ids).lower() == 'all':
            self.read_ids = list(self.tid_set)
        if not hasattr(self.read_ids, '__iter__'):
            self.read_ids = [read_ids]


        f = open(self.filepath, 'rb')
        f.seek(self.data_address)

        # read everything in 8bits, grabs time stamps, get_valid_tids_and_fs called on init has already has ids
        self.t_stamps_256 = self._e_bit_reads[3::4]

        # read again, but in 16 bit chunks, grab messages
        f.seek(self.data_address + 1)

        self.voltage_messages = np.fromfile(f, '>u2')[::2]
        self._merge_coarse_and_fine_clocks() # this generates self.time_array

        for read_id in self.read_ids:
            assert read_id in self.tid_set, "Transmitter %i is not a valid transmitter id" % read_id
            self.tid_raw_data_time_dict[read_id]['data'] = self.voltage_messages[self.transmitter_id_bytes == read_id] * self.micro_volt_div
            self.tid_raw_data_time_dict[read_id]['time'] = self.time_array[self.transmitter_id_bytes == read_id]

        # remove bad messages
        self._correct_bad_messages()

        if auto_glitch_removal:
            self.glitch_removal(tactic='mad')

        if auto_resampling:
            self.correct_sampling_frequency()
            # there should now be no nans surviving here!

        if auto_filter:
            self.highpass_filter()

        if scale_to_mode_std: # not normally called
            self.standardise_to_mode_stddev()


    def highpass_filter(self, cutoff_hz = 1):
        '''
        Implements high pass digital butterworth filter, order 2.

        Args:
            cutoff_hz: default is 1hz
        '''
        for read_id in self.read_ids:
            fs = self.tid_to_fs_dict[read_id]
            nyq = 0.5 * fs
            cutoff_decimal = cutoff_hz/nyq

            logging.info('Highpassfiltering, tid = '+str(read_id)+' fs: ' + str(fs) + ' at '+ str(cutoff_hz)+ ' Hz')
            data = self.tid_data_time_dict[read_id]['data']
            data = data - np.mean(data)    # remove mean to try and reduce any filtering artifacts
            b, a = signal.butter(2, cutoff_decimal, 'highpass', analog=False)
            filtered_data = signal.filtfilt(b, a, data)

            self.tid_data_time_dict[read_id]['data'] = filtered_data

    def standardise_to_mode_stddev(self, stdtw = 5, std_sigfigs = 2):
        '''
        Calculates mode std dev and divides by it.

        Args:
            stdtw: time period over which to calculate std deviation
            std_sigfigs: n signfigs to round to

        '''
        for read_id in self.read_ids:
            fs = self.tid_to_fs_dict[read_id]
            data = self.tid_data_time_dict[read_id]['data']

            logging.info('Standardising to mode std dev, tid = '+str(read_id))

            reshaped = np.reshape(data, (int(3600/stdtw), int(stdtw*fs)))
            #std_vector = self.round_to_sigfigs(np.std(reshaped, axis = 1), sigfigs=std_sigfigs)
            std_vector = np.round(np.std(data, axis = 1), 0)
            std_vector = std_vector[std_vector != 0]
            if std_vector.shape[0] > 0:
                mode_std = stats.mode(std_vector)[0] # can be zero if there is big signal loss
                scaled = np.divide(data, mode_std)
                self.tid_data_time_dict[read_id]['data'] = scaled
                logging.debug(str(mode_std)+' is mode std of trace split into '+ str(stdtw)+ ' second chunks')
            elif std_vector.shape[0] == 0:
                self.tid_data_time_dict[read_id]['data'] =  None
                logging.error(' File std is all 0, changed data to be None')

    @staticmethod
    def round_to_sigfigs(x, sigfigs):
        """
        N.B Stolen from stack overflow:
        http://stackoverflow.com/questions/18915378/rounding-to-significant-figures-in-numpy

        Rounds the value(s) in x to the number of significant figures in sigfigs.

        Restrictions:
        sigfigs must be an integer type and store a positive value.
        x must be a real value or an array like object containing only real values.
        """
        #The following constant was computed in maxima 5.35.1 using 64 bigfloat digits of precision
        __logBase10of2 = 3.010299956639811952137388947244930267681898814621085413104274611e-1
        if not ( type(sigfigs) is int or np.issubdtype(sigfigs, np.integer)):
            raise TypeError( "RoundToSigFigs: sigfigs must be an integer." )

        if not np.all(np.isreal( x )):
            raise TypeError( "RoundToSigFigs: all x must be real." )

        if sigfigs <= 0:
            raise ValueError( "RoundtoSigFigs: sigfigs must be positive." )

        mantissas, binaryExponents = np.frexp(x)

        decimalExponents = __logBase10of2 * binaryExponents
        intParts = np.floor(decimalExponents)

        mantissas *= 10.0**(decimalExponents - intParts)

        return np.around(mantissas, decimals=sigfigs - 1 ) * 10.0**intParts

    def _correct_bad_messages(self):
        '''
        - okay so we have 128hz as the clock...
        - fs / clock_rate is the n_messages between clocks
        - 256 / n_messages is the thing we are diving to get the residuals
        '''
        for tid in self.read_ids:
            transmitter_timestamps = self.t_stamps_256[self.transmitter_id_bytes == tid]
            fs = self.tid_to_fs_dict[tid]
            n_messages = fs/128 # 128 is clock
            expected_interval = 256/n_messages # 256 is bits (if 512hz fs this is 64)
            timestamp_moduli = transmitter_timestamps % expected_interval

            # now get params for reshaping...
            n_rows = int(fs*4)
            #n_rows = 2000
            n_fullcols = int(timestamp_moduli.size//n_rows)
            n_extra_stamps = timestamp_moduli.shape[0] - (n_rows*n_fullcols)
            end_moduli = timestamp_moduli[-n_extra_stamps:]
            if n_extra_stamps:
                reshaped_moduli = np.reshape(timestamp_moduli[:-n_extra_stamps], (n_rows, n_fullcols), order = 'F')
                # order F reshaped in a "fortran manner, first axis changing fastest", calculating down the columns here
                end_mean= ss.circmean(end_moduli, high = expected_interval)
                end_moduli_corrected = (end_moduli - end_mean)
                mean_vector = ss.circmean(reshaped_moduli, high=expected_interval, axis=0)
                moduli_array_corrected = (reshaped_moduli - mean_vector)
                drift_corrected_timestamp_moduli = np.concatenate([np.ravel(moduli_array_corrected, order = 'F'), end_moduli_corrected])

            elif n_extra_stamps == 0: # can be reshaped exactly
                reshaped_moduli = np.reshape(timestamp_moduli, (n_rows, n_fullcols), order = 'F')
                mean_vector = ss.circmean(reshaped_moduli, high=expected_interval, axis=0)
                moduli_array_corrected = (reshaped_moduli - mean_vector)
                drift_corrected_timestamp_moduli = np.ravel(moduli_array_corrected, order = 'F')

            drift_corrected_timestamp_moduli = np.absolute(drift_corrected_timestamp_moduli)
            self.drift_corrected_timestamp_moduli = drift_corrected_timestamp_moduli

            bad_message_locs = np.where(np.logical_and(drift_corrected_timestamp_moduli > 9,
                                                       drift_corrected_timestamp_moduli < (expected_interval-9)))[0]

            self.tid_data_time_dict[tid]['data'] = np.delete(self.tid_raw_data_time_dict[tid]['data'], bad_message_locs)
            self.tid_data_time_dict[tid]['time'] = np.delete(self.tid_raw_data_time_dict[tid]['time'], bad_message_locs)
            logging.debug('Tid ' +str(tid)+ ': Detected '+ str(len(bad_message_locs)) + ' bad messages out of '+ str(self.tid_raw_data_time_dict[tid]['data'].shape[0])
                       + ' Remaining : '+str(self.tid_data_time_dict[tid]['data'].shape[0]))

            if len(bad_message_locs) > 0.5*self.tid_raw_data_time_dict[tid]['data'].shape[0]:
                logging.error(' >half messages detected as bad messages. Probably change fs from auto to the correct frequency')

            if self.verbose:
                print ('Tid ' +str(tid)+ ': Detected '+ str(len(bad_message_locs)) + ' bad messages out of '+ str(self.tid_raw_data_time_dict[tid]['data'].shape[0])
                       + ' Remaining : '+str(self.tid_data_time_dict[tid]['data'].shape[0]))
            #if len(bad_message_locs) > 0.5*self.tid_raw_data_time_dict[tid]['data'].shape[0]:
            #    print('WARNING: >half messages detected as bad messages. Probably change fs from auto to the correct frequency')







# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'library_window.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_LibraryManagement(object):
    def setupUi(self, LibraryManagement):
        LibraryManagement.setObjectName("LibraryManagement")
        LibraryManagement.resize(1050, 833)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LibraryManagement.sizePolicy().hasHeightForWidth())
        LibraryManagement.setSizePolicy(sizePolicy)
        LibraryManagement.setMinimumSize(QtCore.QSize(1050, 833))
        LibraryManagement.setMaximumSize(QtCore.QSize(2000, 833))
        font = QtGui.QFont()
        font.setKerning(True)
        LibraryManagement.setFont(font)
        LibraryManagement.setSizeGripEnabled(False)
        LibraryManagement.setModal(False)
        self.gridLayout_2 = QtWidgets.QGridLayout(LibraryManagement)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.annotations_display = QtWidgets.QLineEdit(LibraryManagement)
        self.annotations_display.setMinimumSize(QtCore.QSize(500, 0))
        self.annotations_display.setDragEnabled(True)
        self.annotations_display.setReadOnly(True)
        self.annotations_display.setObjectName("annotations_display")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.annotations_display)
        self.select_annotations = QtWidgets.QPushButton(LibraryManagement)
        self.select_annotations.setMinimumSize(QtCore.QSize(200, 0))
        self.select_annotations.setObjectName("select_annotations")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.select_annotations)
        self.select_h5_folder = QtWidgets.QPushButton(LibraryManagement)
        self.select_h5_folder.setMinimumSize(QtCore.QSize(200, 0))
        self.select_h5_folder.setObjectName("select_h5_folder")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.select_h5_folder)
        self.h5_folder_display = QtWidgets.QLineEdit(LibraryManagement)
        self.h5_folder_display.setMinimumSize(QtCore.QSize(500, 0))
        self.h5_folder_display.setObjectName("h5_folder_display")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.h5_folder_display)
        self.gridLayout_2.addLayout(self.formLayout, 1, 0, 1, 5)
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_8 = QtWidgets.QLabel(LibraryManagement)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_8.sizePolicy().hasHeightForWidth())
        self.label_8.setSizePolicy(sizePolicy)
        self.label_8.setObjectName("label_8")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.fs_box = QtWidgets.QLineEdit(LibraryManagement)
        self.fs_box.setMaximumSize(QtCore.QSize(50, 16777215))
        self.fs_box.setObjectName("fs_box")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.fs_box)
        self.gridLayout_2.addLayout(self.formLayout_2, 2, 4, 1, 1)
        self.label_6 = QtWidgets.QLabel(LibraryManagement)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_6.sizePolicy().hasHeightForWidth())
        self.label_6.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Helvetica Neue")
        font.setPointSize(19)
        font.setBold(False)
        font.setWeight(50)
        self.label_6.setFont(font)
        self.label_6.setAlignment(QtCore.Qt.AlignCenter)
        self.label_6.setIndent(-1)
        self.label_6.setObjectName("label_6")
        self.gridLayout_2.addWidget(self.label_6, 3, 0, 1, 5)
        self.label = QtWidgets.QLabel(LibraryManagement)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 5, 0, 1, 5)
        self.library_path_display = QtWidgets.QLineEdit(LibraryManagement)
        self.library_path_display.setMinimumSize(QtCore.QSize(500, 0))
        self.library_path_display.setReadOnly(True)
        self.library_path_display.setObjectName("library_path_display")
        self.gridLayout_2.addWidget(self.library_path_display, 6, 0, 1, 5)
        self.set_library = QtWidgets.QPushButton(LibraryManagement)
        self.set_library.setObjectName("set_library")
        self.gridLayout_2.addWidget(self.set_library, 7, 0, 1, 1)
        self.clear_library = QtWidgets.QPushButton(LibraryManagement)
        self.clear_library.setObjectName("clear_library")
        self.gridLayout_2.addWidget(self.clear_library, 7, 1, 1, 2)
        self.new_library = QtWidgets.QPushButton(LibraryManagement)
        self.new_library.setObjectName("new_library")
        self.gridLayout_2.addWidget(self.new_library, 2, 0, 1, 1)
        self.line_2 = QtWidgets.QFrame(LibraryManagement)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_2.addWidget(self.line_2, 4, 0, 1, 5)
        self.add_to_library = QtWidgets.QPushButton(LibraryManagement)
        self.add_to_library.setObjectName("add_to_library")
        self.gridLayout_2.addWidget(self.add_to_library, 2, 1, 1, 2)
        self.label_5 = QtWidgets.QLabel(LibraryManagement)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Helvetica Neue")
        font.setPointSize(19)
        font.setBold(False)
        font.setWeight(50)
        self.label_5.setFont(font)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setIndent(-1)
        self.label_5.setObjectName("label_5")
        self.gridLayout_2.addWidget(self.label_5, 0, 0, 1, 5)
        self.add_labels = QtWidgets.QPushButton(LibraryManagement)
        self.add_labels.setObjectName("add_labels")
        self.gridLayout_2.addWidget(self.add_labels, 8, 0, 1, 2)
        self.use_peaks = QtWidgets.QCheckBox(LibraryManagement)
        self.use_peaks.setObjectName("use_peaks")
        self.gridLayout_2.addWidget(self.use_peaks, 10, 0, 1, 5)
        self.progressBar_lable_above1 = QtWidgets.QLabel(LibraryManagement)
        self.progressBar_lable_above1.setObjectName("progressBar_lable_above1")
        self.gridLayout_2.addWidget(self.progressBar_lable_above1, 13, 0, 1, 5)
        self.line_3 = QtWidgets.QFrame(LibraryManagement)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout_2.addWidget(self.line_3, 12, 0, 1, 5)
        self.label_4 = QtWidgets.QLabel(LibraryManagement)
        self.label_4.setObjectName("label_4")
        self.gridLayout_2.addWidget(self.label_4, 8, 2, 1, 3)
        self.add_features = QtWidgets.QPushButton(LibraryManagement)
        self.add_features.setObjectName("add_features")
        self.gridLayout_2.addWidget(self.add_features, 9, 0, 1, 2)
        self.label_7 = QtWidgets.QLabel(LibraryManagement)
        self.label_7.setObjectName("label_7")
        self.gridLayout_2.addWidget(self.label_7, 9, 2, 1, 3)
        self.overwrite_box = QtWidgets.QCheckBox(LibraryManagement)
        self.overwrite_box.setObjectName("overwrite_box")
        self.gridLayout_2.addWidget(self.overwrite_box, 11, 0, 1, 5)
        self.progressBar = QtWidgets.QProgressBar(LibraryManagement)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setTextVisible(True)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout_2.addWidget(self.progressBar, 14, 0, 1, 5)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.gridLayout.setContentsMargins(0, -1, -1, -1)
        self.gridLayout.setHorizontalSpacing(12)
        self.gridLayout.setObjectName("gridLayout")
        self.label_3 = QtWidgets.QLabel(LibraryManagement)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 0, 3, 1, 1)
        self.chunk_length = QtWidgets.QLineEdit(LibraryManagement)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.chunk_length.sizePolicy().hasHeightForWidth())
        self.chunk_length.setSizePolicy(sizePolicy)
        self.chunk_length.setMinimumSize(QtCore.QSize(50, 0))
        self.chunk_length.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.chunk_length.setSizeIncrement(QtCore.QSize(0, 0))
        self.chunk_length.setBaseSize(QtCore.QSize(0, 0))
        self.chunk_length.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.chunk_length.setObjectName("chunk_length")
        self.gridLayout.addWidget(self.chunk_length, 0, 2, 1, 1)
        self.label_2 = QtWidgets.QLabel(LibraryManagement)
        self.label_2.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setMinimumSize(QtCore.QSize(140, 0))
        self.label_2.setMaximumSize(QtCore.QSize(100, 16777215))
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 1, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 7, 3, 1, 1)
        self.progressBar_label_above2 = QtWidgets.QLabel(LibraryManagement)
        self.progressBar_label_above2.setGeometry(QtCore.QRect(30, 580, 850, 16))
        self.progressBar_label_above2.setText("")
        self.progressBar_label_above2.setObjectName("progressBar_label_above2")

        self.retranslateUi(LibraryManagement)
        QtCore.QMetaObject.connectSlotsByName(LibraryManagement)

    def retranslateUi(self, LibraryManagement):
        _translate = QtCore.QCoreApplication.translate
        LibraryManagement.setWindowTitle(_translate("LibraryManagement", "Dialog"))
        self.select_annotations.setText(_translate("LibraryManagement", "Select annotations"))
        self.select_h5_folder.setText(_translate("LibraryManagement", "Select H5 folder"))
        self.label_8.setToolTip(_translate("LibraryManagement", "Sampling Frequency"))
        self.label_8.setText(_translate("LibraryManagement", "Fs (Hz)"))
        self.fs_box.setText(_translate("LibraryManagement", "512"))
        self.label_6.setText(_translate("LibraryManagement", "Add Features and Labels"))
        self.label.setText(_translate("LibraryManagement", "Library path:"))
        self.set_library.setText(_translate("LibraryManagement", "Set library"))
        self.clear_library.setText(_translate("LibraryManagement", "Clear library path "))
        self.new_library.setText(_translate("LibraryManagement", "New library"))
        self.add_to_library.setText(_translate("LibraryManagement", "Add to library"))
        self.label_5.setText(_translate("LibraryManagement", "Make Library"))
        self.add_labels.setText(_translate("LibraryManagement", "Add labels to library"))
        self.use_peaks.setText(_translate("LibraryManagement", "Use peaks and valleys features, slower!"))
        self.progressBar_lable_above1.setText(_translate("LibraryManagement", "Progress"))
        self.label_4.setText(_translate("LibraryManagement", "Quickly add labels for looking at how it chunks up in main gui."))
        self.add_features.setText(_translate("LibraryManagement", "Add features to library"))
        self.label_7.setText(_translate("LibraryManagement", "Warning, this will be slow! Labels added automatically."))
        self.overwrite_box.setText(_translate("LibraryManagement", "Overwrite previously calculated features in library"))
        self.label_3.setText(_translate("LibraryManagement", "s"))
        self.chunk_length.setText(_translate("LibraryManagement", "5"))
        self.label_2.setText(_translate("LibraryManagement", "Chunk length"))




import sys
import os
import multiprocessing
import shutil
import traceback
import time
import logging

import h5py
import numpy as np
import pandas as pd
from scipy import signal, stats

from .ndfconverter import NdfFile
from .h5loader import H5File
from .feature_extractor import FeatureExtractor

if sys.version_info < (3,):
    range = xrange

import logging

#from make_pdfs import plot_traces_hdf5, plot_traces
class DataHandler():
    '''
    Class to handle all ingesting of data and outputing it in a format for the Classifier Handler

    TODO:
     -  pass in dates for predictions, and print out files you skipped is not transmitter there

    To think about
    - enable feature extraction to be better handled
    - remeber when defining maxshape = (None, data_array.shape[1])) - to allow h5 to be resized


    import multiprocessing as mp
    import time

    def foo_pool(x):
        time.sleep(2)
        return x*x

    result_list = []
    def log_result(result):
        # This is called whenever foo_pool(i) returns a result.
        # result_list is modified only by the main process, not the pool workers.
        result_list.append(result)

    def apply_async_with_callback():
        pool = mp.Pool(4)
        for i in range(10):
        pool.imap(foo_pool, args = (i, ), callback = log_result)
        pool.close()
        pool.join()
        print(result_list)

apply_async_with_callback()
    '''

    def __init__(self, logpath = os.getcwd()):

        self.parallel_savedir = None
        self.parrallel_flag_pred = False

    def add_labels_to_seizure_library(self, library_path, overwrite, timewindow):
        '''

        Called by self.add_features_seizure_library()

        '''
        logging.info('Adding labels to '+ library_path)
        # todo check timewindows go into n seconds precisely
        with h5py.File(library_path, 'r+') as f:
            seizure_datasets = [f[group] for group in list(f.keys())]

            for i, group in enumerate(seizure_datasets):
                # First check if there are labels already there. If overwrite, go to except block
                try:
                    if not overwrite:
                        features = group['labels']
                        logging.info(str(group)+' already has labels, skipping')
                    if overwrite:
                        raise
                except:
                    try: # attempt to delete, will throw error if group doesn't exist.
                        del group['labels']
                        del group.attrs['chunked_annotation']

                        logging.debug('Deleted old labels')
                    except:
                        pass
                    # make labels dataset here:
                    # first work out the would-be shape of the data array
                    data = group['data'][:]
                    fs = group.attrs['fs']
                    n_traces = int(data.shape[0] / (fs * timewindow))
                    #dp_lost =  int(data.shape[0] % (fs * timewindow))
                    # dp per chunk is then int(fs * timewindow))
                    #print(n_traces)
                    labels = np.zeros(n_traces)
                    #print(group.attrs['precise_annotation'])
                    group.attrs['chunk_length'] = timewindow
                    for row in group.attrs['precise_annotation']:
                        start_i = int(np.floor(row[0]/timewindow))
                        end_i   = int(np.ceil(row[1]/timewindow))
                        labels[start_i:end_i] = 1

                        try:
                            group.attrs['chunked_annotation'] = np.vstack([group.attrs['chunked_annotation'],
                                                                           np.array([(start_i*timewindow,end_i*timewindow)]) ])
                        except KeyError:
                            group.attrs['chunked_annotation'] = np.array([(start_i*timewindow,end_i*timewindow)])
                    ###
                    #print(group.attrs['chunked_annotation'])


                    group.create_dataset('labels', data = labels, compression = "gzip", dtype = 'i2', chunks = labels.shape)

    def add_features_seizure_library(self, library_path, overwrite = False, run_peaks = True, timewindow = 5):
        '''
        Inputs:
        - overwrite: if false, will only add labels and features to seizures in the library that do not already
          have accompanying labels and features.

        This is painfully slow at the moment, not accessing hdf5 groups in parallel.
        Both adding features to library and prediction h5 file now are expecting unchunked data. This method
        will make labels too though
        '''
        logging.info('First adding labels to ' + library_path)
        print('Adding labels first...', end=' ')
        self.add_labels_to_seizure_library(library_path, overwrite, timewindow)
        print('Done')

        logging.info('Adding features to ' + library_path)
        with h5py.File(library_path, 'r+') as f:
            seizure_datasets = [f[group] for group in list(f.keys())]

            l = len(seizure_datasets)-1
            self.printProgress(0,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
            for i, group in enumerate(seizure_datasets):
                logging.debug('Loading group: '+ str(group))

                # First check if there are features already there. If overwrite, go to except block
                try:
                    if not overwrite:
                        features = group['features']
                        logging.info(str(group)+' already has features, skipping')
                    if overwrite:
                        raise
                except:
                    data_array = group['data'][:]
                    data_array = self._make_array_from_data(data_array, fs = group.attrs['fs'], timewindow = timewindow)
                    assert len(data_array.shape) > 1

                    if data_array is not None:
                        extractor = FeatureExtractor(data_array, fs = group.attrs['fs'], run_peakdet = run_peaks)
                        features = extractor.feature_array
                        try: # attempt to delete, will throw error if group doesn't exist.
                            del group['features']
                            logging.debug('Deleted old features')
                        except:
                            pass
                        group.create_dataset('features', data = features, compression = 'gzip', dtype = 'f4')
                        group.attrs['col_names'] = np.array(extractor.col_labels).astype('|S9')
                        # here add feature titles to the dataset attrs?
                        logging.info('Added features to ' + str(group) + ', shape:' + str(features.shape))
                        self.printProgress(i,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
                    # Data_array is none
                    else:
                        logging.error("Didn't add features to file: "+str(group))
                        return 0

    def add_predicition_features_to_h5_file(self, h5_file_path, timewindow = 5, run_peakdet = True):
        '''
        Currently assuming only one transmitter per h5 file

        Could eventually (now only one tid) have the structure of:

        M1445612etc / tid1
                    / tid2
                    / tid3  / data
                            / time
        '''
        if self.parrallel_flag_pred:
            run_peakdet = self.run_pkdet
            timewindow  = self.twindow

        with h5py.File(h5_file_path, 'r+') as f:

            mcodes = [f[group] for group in list(f.keys())]
            assert len(mcodes) == 1 # assuming one transmitter per file
            mcode = mcodes[0]

            tids = [mcode[tid] for tid in list(mcode.keys())]

            for tid in tids: # loop through groups which are tids for predicition h5s
                data_array = tid['data'][:]

                if data_array is not None:

                    logging.debug('Reshaping data from '+str(data_array.shape)+ ' using '+str(tid.attrs['fs']) +' fs' +
                                  ' and timewindow of '+ str(timewindow)  )
                    data_array = self._make_array_from_data(data_array, fs = tid.attrs['fs'], timewindow = timewindow)
                    try:
                        assert data_array.shape[0] == int(3600/timewindow)
                    except:
                        print('Warning: Data file does not contain, full data: ' + str(os.path.basename(h5_file_path)) + str(data_array.shape))
                    #return data_array
                    extractor = FeatureExtractor(data_array, tid.attrs['fs'], run_peakdet = run_peakdet)
                    features = extractor.feature_array

                    try:
                        del tid['features']
                        logging.debug('Deleted old features')
                    except:
                        pass
                    tid.create_dataset('features', data = features, compression = 'gzip')
                    tid.attrs['col_names'] = np.array(extractor.col_labels).astype('|S9')

                    logging.debug('Added features to ' + str(os.path.basename(h5_file_path)) + ', shape:' + str(features.shape))

                elif data_array is None:
                    logging.error('File has None for data - perhaps all the same')
                    logging.error("Didn't add features to group: "+ str(tid))

                    return 0

    def parallel_add_prediction_features(self, h5py_folder, n_cores = -1, run_peakdet = False, timewindow = 5):
        '''
        # NEED TO ADD SETTINGS HERE FOR THE TIMEWINDOW ETC

        '''
        self.parrallel_flag_pred = True
        self.run_pkdet = run_peakdet
        self.twindow = timewindow
        files_to_add_features = [os.path.join(h5py_folder, fname) for fname in os.listdir(h5py_folder) if fname.startswith('M')] #switch to self.fullpath_listdir()
        if n_cores == -1:
            n_cores = multiprocessing.cpu_count()

        pool = multiprocessing.Pool(n_cores)
        l = int(len(files_to_add_features))
        print( ' Adding features to '+str(l)+ ' hours in '+ h5py_folder)
        self.printProgress(0,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
        for i, _ in enumerate(pool.imap(self.add_predicition_features_to_h5_file, files_to_add_features), 1):
            self.printProgress(i,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)

        #pool.map(self.add_predicition_features_to_h5_file, files_to_add_features)
        pool.close()
        pool.join()
        self.parrallel_flag_pred = False
    def get_annotations_from_df_datadir_matches(self, df, file_dir):
        '''
        This function matches the entries in a dataframe with files in a directory

        Returns: list of annotations stored in a list
        '''

        abs_filenames = [os.path.join(file_dir, f) for f in os.listdir(file_dir)]
        data_filenames = [f for f in os.listdir(file_dir) if f.startswith('M')]
        mcodes = [os.path.split(f)[1][:11] for f in os.listdir(file_dir) if f.startswith('M')]
        n_files = len(data_filenames)

        # now loop through matching the tid to datafile in the annotations
        df.columns = [label.lower() for label in df.columns]
        df.columns  = [label.strip(' ') for label in df.columns]

        reference_count = 0
        annotation_dicts = []
        for row in df.iterrows():
            # annotation name is bad, but will ultimately be the library h5 dataset name
            annotation_name = str(row[1]['filename']).split('.')[0]+'_tid_'+str(int(row[1]['transmitter']))
            for datafile in data_filenames:
                if datafile.startswith(annotation_name.split('_')[0]):
                    start = row[1]['start']
                    end = row[1]['end']
                    annotation_dicts.append({'fname': os.path.join(file_dir, datafile),
                                             'start': start,
                                             'end': end,
                                             'dataset_name': annotation_name,
                                             'tid':int(row[1]['transmitter'])})
                    reference_count += 1

        print('Of the '+str(n_files)+' ndfs in directory, '+str(reference_count)+' references to seizures were found in the passed dataframe')
        return annotation_dicts

    def make_seizure_library(self, df, file_dir,
                             timewindow = 5,
                             seizure_library_name = 'seizure_library',
                             fs = 'auto',
                             verbose = False,
                             overwrite = False,
                             scale_and_filter = False):
        '''
        Args:

            df : pandas dataframe. Column titles need to be "filename", "start","end", "transmitter"
            file_dir: path to converted h5, or ndf directory, that contains files referenced in
                      the dataframe
            timewindow: size to chunk the data up with
            seizure_library_name: path and name of the seizure lib.
            fs: default is auto, but use freq in hz to sidestep the auto dectection
            verbose: Flag, print or not.
            - scale and filter should not be used (for simply keeping in original y units)

        Returns:
            Makes a Seizure library file

        WARNING: The annotation will be incorrect based on the time-window coarseness and the chunk that is chosen!
        Currently finding the start by start/timewindom -- end/timewindow


        TODO:
        -  How to handle files that don't have seiures, but we want to include
        -  Not sure what is going on when there are no seizures, need to have this functionality though.

        '''

        logging.info('Datahandler - creating SeizureLibrary')
        try:
            annotation_dicts = self.get_annotations_from_df_datadir_matches(df, file_dir)
        except:
            print("Error getting annotations from your file. Please ensure columns are named: 'filename', 'transmitter','start','end'")
            annotation_dicts = self.get_annotations_from_df_datadir_matches(df, file_dir)
        # annotations_dicts is a list of dicts with... e.g 'dataset_name': 'M1445443776_tid_9',
        # 'end': 2731.0, 'fname': 'all_ndfs/M1445443776.ndf', 'start': 2688.0,' tid': 9

        h5code = 'w' if overwrite else 'x'
        try:
            if not '/' in seizure_library_name or not "\\" in seizure_library_name:
                seizure_library_path = os.path.split(file_dir)[0]+seizure_library_name.strip('.h5')+'.h5'
            seizure_library_path = seizure_library_name.strip('.h5')+'.h5'
            print('Creating seizure library: '+ seizure_library_path)
            logging.info('Creating seizure library: '+ seizure_library_path)
            h5file = h5py.File(seizure_library_path, h5code)
            h5file.attrs['fs'] = fs
            h5file.attrs['timewindow'] = timewindow
            h5file.close()

        except Exception:
            print ('Error: Seizure library file exists! Delete it or set "overwrite" to True')
            exc_type, exc_value, exc_traceback = sys.exc_info()
            print (traceback.print_exception(exc_type, exc_value, exc_traceback))
            return 0

        # now populate to seizure lib with data, time and labels
        # make a list
        l = len(annotation_dicts)-1
        self.printProgress(0,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
        for i, annotation in enumerate(annotation_dicts):
            self._populate_seizure_library(annotation,
                                           fs,
                                           timewindow,
                                           seizure_library_path,
                                           verbose = verbose,
                                           scale_and_filter = scale_and_filter)
            self.printProgress(i,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)

    def append_to_seizure_library(self, df, file_dir, seizure_library_path,
                                  timewindow = 5,
                                  fs = 'auto',
                                  verbose = False,
                                  overwrite = False,
                                  scale_and_filter = False):
        '''
        Args:

            df : pandas dataframe. Column titles need to be "name", "start","end", "transmitter"
            file_dir: path to converted h5, or ndf directory, that contains files referenced in
                      the dataframe
            timewindow: size to chunk the data up with
            seizure_library_path: path and name of the seizure lib.
            fs: default is auto, but use freq in hz to sidestep the auto dectection
            verbose: Flag, print or not.

        Returns:
            Appends to a Seizure library file

        WARNING: The annotation will be incorrect based on the time-window coarseness and the chunk that is chosen!
        Currently finding the start by start/timewindom -- end/timewindow


        TODO:
        -  add documentaiton of the loading settings
        -  How to handle files that don't have seiures, but we want to include (done, put to 00's)
        -  Not sure what is going on when there are no seizures, need to have this functionality though.

        '''

        logging.info('Appending to seizure library')
        annotation_dicts = self.get_annotations_from_df_datadir_matches(df, file_dir)
        # annotations_dicts is a list of dicts with... e.g 'dataset_name': 'M1445443776_tid_9',
        # 'end': 2731.0, 'fname': 'all_ndfs/M1445443776.ndf', 'start': 2688.0,' tid': 9

        # now add to to seizure lib with data, time and labels
        l = len(annotation_dicts)-1
        logging.info('Datahandler - creating SeizureLibrary')
        self.printProgress(0,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
        for i, annotation in enumerate(annotation_dicts):
            self._populate_seizure_library(annotation, fs, timewindow, seizure_library_path, verbose = verbose, scale_and_filter = scale_and_filter)
            self.printProgress(i,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)


    def _populate_seizure_library(self, annotation, fs,
                                  timewindow,
                                  seizure_library_path,
                                  verbose = False,
                                  scale_and_filter = False):


        logging.debug('Adding '+str(annotation['fname']))
        if annotation['fname'].endswith('.ndf'):
            file_obj = NdfFile(annotation['fname'],fs = fs)
            file_obj.load(annotation['tid'], scale_to_mode_std= scale_and_filter)
        elif annotation['fname'].endswith('.h5'):
            file_obj = H5File(annotation['fname'])
        else:
            print('ERROR: Unrecognised file-type')

        # add in filtering and scaling here
        #data_array = self._make_array_from_data(file_obj[annotation['tid']]['data'], fs, timewindow)
        data_array = file_obj[annotation['tid']]['data'] # just 1D at the moment!

        # use the start and end times to make labels
        #labels = np.zeros(shape = (data_array.shape[0]))
        #start_i = int(np.floor(annotation['start']/timewindow))
        #end_i   = int(np.ceil(annotation['end']/timewindow))

        #print('  ')
        #print(annotation['start'], ': ', start_i*timewindow)
        #print(annotation['end'], ':', end_i*timewindow)

        with h5py.File(seizure_library_path, 'r+') as f:
            if annotation['dataset_name'] in f.keys():
                logging.info(str(annotation['dataset_name'])+' has more than one seizure!')
                #labels =  f[annotation['dataset_name']+'/labels']
                # todo wrap this in try, not possible to have end earlier than start of the seizure, will throw reverse selection error
                #labels[start_i:end_i] = 1
                #print('more than one seizure')

                #f[annotation['dataset_name']].attrs['chunked_annotation'] = np.vstack(
                #    [f[annotation['dataset_name']].attrs['chunked_annotation'], np.array([(120,160)]) ])

                #print(f[annotation['dataset_name']].attrs['chunked_annotation'])
                f[annotation['dataset_name']].attrs['precise_annotation'] = np.vstack(
                    [f[annotation['dataset_name']].attrs['precise_annotation'], np.array([(annotation['start'],annotation['end'])])])
                #print((f[annotation['dataset_name']].attrs['precise_annotation']))
                #print(dict(f[annotation['dataset_name']].attrs))

            else:
                group = f.create_group(annotation['dataset_name'])
                group.attrs['tid'] = annotation['tid']
                group.attrs['fs']  = fs
                group.attrs['scaled_and_filtered'] = scale_and_filter
                #group.attrs['chunked_annotation'] = np.array([(100,160)])
                group.attrs['precise_annotation'] = np.array([(annotation['start'],annotation['end'])])
                group.create_dataset('data', data = data_array, compression = "gzip", dtype='f4', chunks = data_array.shape)
                #labels[start_i:end_i] = 1 # indexing is fine, dont need to convert to array
                #group.create_dataset('labels', data = labels, compression = "gzip", dtype = 'i2', chunks = labels.shape)
            f.close()

    @staticmethod
    def _make_array_from_data(data, fs, timewindow):
        '''
        What ordering are we looking for here?
        '''
        n_traces = int(data.shape[0] / (fs * timewindow))
        dp_lost =  int(data.shape[0] % (fs * timewindow))
        if dp_lost > 0:
            data_array = np.reshape(data[:-dp_lost], newshape = (n_traces, int(fs * timewindow)))
        else:
            data_array = np.reshape(data, newshape = (n_traces, int(fs * timewindow)))
        return data_array

    @staticmethod
    def fullpath_listdir(d):
        ''' returns full filepath,  excludes hidden files, starting with .'''
        return [os.path.join(d, f) for f in os.listdir(d) if not f.startswith('.')]

    def convert_ndf_directory_to_h5(self, ndf_dir, tids = 'all', save_dir  = 'same_level', n_cores = -1, fs = 'auto'):
        """

        Args:
            ndf_dir: Directory to convert
            tids: transmitter ids to convert. Default is 'all'
            save_dir: optional save directory, will default to appending convertered_h5s after current ndf
            n_cores: number of cores to use
            fs :  'auto' or frequency in hz

        ndfs conversion seem to be pretty buggy...

        """

        self.fs_for_parallel_conversion = fs
        files = [f for f in self.fullpath_listdir(ndf_dir) if f.endswith('.ndf')]

        # check ids
        ndf = NdfFile(files[0])
        if not tids == 'all':
            if not hasattr(tids, '__iter__'):
                tids = [tids]

        self.tids_for_parallel_conversion = tids
        print (str(len(files))+' Files for conversion. Transmitters: '+ str(self.tids_for_parallel_conversion))

        # set n_cores
        if n_cores == -1:
            n_cores = multiprocessing.cpu_count()

        # Make save directory
        if save_dir  == 'same_level':
            save_dir = ndf_dir+'_converted_h5s'
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        self.savedir_for_parallel_conversion = save_dir

        pool = multiprocessing.Pool(n_cores)
        l = len(files)
        self.printProgress(0,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
        for i, _ in enumerate(pool.imap(self.convert_ndf, files), 1):
            self.printProgress(i,l, prefix = 'Progress:', suffix = 'Complete', barLength = 50)
        pool.close()
        pool.join()

    def convert_ndf(self, filename):

        savedir = self.savedir_for_parallel_conversion
        tids = self.tids_for_parallel_conversion
        fs = self.fs_for_parallel_conversion

        # convert m name
        mname = os.path.split(filename)[1]
        tstamp = float(mname.strip('M').split('.')[0])
        ndf_time = '_'+str(pd.Timestamp.fromtimestamp(tstamp)).replace(':', '_')
        ndf_time =  ndf_time.replace(' ', '_')
        start = time.time()
        try:
            ndf = NdfFile(filename, fs = fs)
            if set(tids).issubset(ndf.tid_set) or tids == 'all':
                ndf.load(tids)
                abs_savename = os.path.join(savedir, os.path.split(filename)[-1][:-4]+ndf_time+' tids_'+str(ndf.read_ids))
                ndf.save(save_file_name= abs_savename)
            else:
                logging.warning('Not all read tids:'+str(tids) +' were valid for '+str(os.path.split(filename)[1])+' skipping!')

        except Exception:
            print('Something unexpected went wrong loading '+str(tids)+' from '+mname+' :')
            #print('Valid ids are:'+str(ndf.tid_set))
            exc_type, exc_value, exc_traceback = sys.exc_info()
            print (traceback.print_exception(exc_type, exc_value,exc_traceback))
        return 0 # don't think i actually this
    # Print iterations progress

    def printProgress (self, iteration, total, prefix = '', suffix = '', decimals = 2, barLength = 100):
        """
        Call in a loop to create terminal progress bar
        @params:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            decimals    - Optional  : number of decimals in percent complete (Int)
            barLength   - Optional  : character length of bar (Int)
        """
        filledLength    = int(round(barLength * iteration / float(total)))
        percents        = round(100.00 * (iteration / float(total)), decimals)
        bar             = '█' * filledLength + '-' * (barLength - filledLength)
        sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),
        sys.stdout.flush()
        if iteration == total:
            sys.stdout.write('\n')
            sys.stdout.flush()
